import json

from django import template
from django.utils.html import format_html
from django.utils.safestring import mark_safe

register = template.Library()


@register.simple_tag(takes_context=True)
def inertia(context, app_id="app"):
    page = context["page__"]
    return format_html(
        "<div id=\"{}\" data-page='{}'></div>", mark_safe(app_id), mark_safe(json.dumps(page))
    )



from django.test import TestCase

from django_inertia import Inertia


class TestInertia(TestCase):
    def setUp(self) -> None:
        super().setUp()
        Inertia().flush_shared()

    def test_script_rendered(self):
        pass

    def test_render(self):
        # check props, components, version, url are here
        pass

    def test_location(self):
        response = Inertia.location("https://github.com")
        self.assertEqual(409, response.status_code)
        self.assertEqual(response.headers.get("X-Inertia-Location"), "https://github.com")

    # @override_settings(INERTIA_ROOT_VIEW="global_test.html")
    # def test_overriding_root_view(self):
    #     assert Inertia().options.get("root_view") == "global_test.html"



from django_inertia import Inertia


def home(request):
    return Inertia.render(request, "Index", {"message": "Hello World"})


def home_extra_data(request):
    return Inertia.render(request, "Index", {"message": "Hello World"}, view_data={
        'app_name': 'Django Inertia',
        'app_version': '1.0'
    })


def view_with_lazy_props(request):
    return Inertia.render(
        request, "Index", {"message": "Hello World", "data": Inertia.lazy(lambda r: "value")}
    )


def view_with_lazy_props_using_request(request):
    return Inertia.render(
        request,
        "Index",
        {"message": "Hello World", "data": Inertia.lazy(lambda r: r.user.first_name)},
    )


def view_with_view_data(request):
    return Inertia.render(
        request,
        "Index",
        {"message": "Hello World", "data": Inertia.lazy(lambda r: r.user.first_name)},
        view_data={'name': Inertia.lazy(lambda r: r.user.first_name)}
    )



import pytest
from django.contrib.auth import get_user_model
from django.test import Client


@pytest.fixture()
def john(db):
    return get_user_model().objects.create(
        username="john", first_name="John", last_name="Doe", email="john.doe@django.org"
    )


@pytest.fixture()
def john_client(john):
    client = Client()
    client.force_login(john)
    return client



import uuid

from django.http import HttpResponse, HttpResponseRedirect
from django.test import RequestFactory, TestCase

from django_inertia.core import Inertia
from django_inertia.middleware import InertiaMiddleware


class TestInertiaMiddleware(TestCase):
    """Test indenpendently middleware flow"""

    def setUp(self) -> None:
        super().setUp()
        self.middleware = InertiaMiddleware(lambda x: HttpResponse())
        self.request = RequestFactory().get("/")
        self.request.session = {}

    def test_initiate_location_if_version_changed(self):
        self.request.headers = {
            "X-Inertia": "true",
            "X-Requested-With": "XMLHttpRequest",
            "X-Inertia-Version": uuid.uuid4(),
        }
        response = self.middleware(self.request)
        assert response.status_code == 409
        assert response.headers.get("X-Inertia-Location", "/")

    def test_redirect_303_for_put_patch_delete_redirect_requests(self):
        middleware = InertiaMiddleware(lambda x: HttpResponseRedirect(redirect_to="/home"))

        request = RequestFactory().put("/users/1")
        request.session = {}
        response = middleware(request)
        assert response.status_code == 303

        request = RequestFactory().patch("/users/1")
        request.session = {}
        response = middleware(request)
        assert response.status_code == 303

        request = RequestFactory().delete("/users/1")
        request.session = {}
        response = middleware(request)
        assert response.status_code == 303

    def test_that_inertia_version_is_appended_if_no_version(self):
        Inertia.version("")
        self.request.headers = {
            "X-Inertia": "true",
            "X-Requested-With": "XMLHttpRequest",
        }
        self.middleware(self.request)
        assert Inertia.get_version() == "1"



from typing import Any, Callable


class LazyProp:
    """Prop that will never be loaded the first time but can be loaded during partial reloads."""

    def __init__(self, callable_arg: Callable):
        self.callable = callable_arg

    def __call__(self, *args):
        return self.callable(*args)


class StaticProp:
    """Prop that will always be loaded even during partial reloads when not present in only.
    Can be useful to always share errors for example.
    As they are always loaded, a callable is not necessary so this takes a static value.
    """

    def __init__(self, value: Any):
        self.value = value

    def __call__(self, *args):
        return self.value



import pytest
from django.urls import reverse_lazy

from django_inertia import Inertia


def share_auth(request):
    if request.user.is_authenticated:
        return {"first_name": request.user.first_name}
    else:
        return False


@pytest.mark.django_db
class TestRequests:
    def test_can_share_user_globally(self, john_client, client):
        # make an unauthenticated request
        response = client.get(reverse_lazy("home"))
        output_props = response.context.get("page__").get("props")
        assert output_props.get("message") == "Hello World"
        assert not output_props.get("user")
        # make a request authenticated as John
        Inertia.share("user", share_auth)
        response = john_client.get(reverse_lazy("home"))
        output_props = response.context.get("page__").get("props")
        assert output_props.get("message") == "Hello World"
        assert output_props.get("user") == {"first_name": "John"}

        Inertia().flush_shared()

    def test_template_contains_view_data(self, client):
        response = client.get(reverse_lazy("home.extra_data"))
        output_props = response.context.get("page__").get("props")
        assert output_props.get("message") == "Hello World"
        assert response.context['app_name'] == "Django Inertia"
        assert response.context['app_version'] == "1.0"



from django.urls import path
from django.views.generic import TemplateView

from . import views

urlpatterns = [
    path("/", TemplateView.as_view(template_name="app.html"), name="test"),
    path("home/", views.home, name="home"),
    path("home/app-data", views.home_extra_data, name="home.extra_data"),
]



import json

from django.template import Context, Template
from django.test import TestCase


class TestInertiaTemplateTag(TestCase):
    def test_rendering_inertia_tag(self):
        rendered = Template("{% load inertia_tags %} {% inertia %}").render(
            Context(
                {
                    "page__": {
                        "component": "Index",
                        "url": "/",
                        "props": {"message": "Hello"},
                        "version": "1",
                    }
                }
            )
        )
        assert 'id="app"' in rendered
        assert "component" in rendered
        assert "props" in rendered
        assert "version" in rendered
        assert "url" in rendered
        assert "message" in rendered

    def test_can_change_app_id(self):
        rendered = Template("{% load inertia_tags %} {% inertia 'my_app' %}").render(
            Context(
                {
                    "page__": {
                        "component": "Index",
                        "url": "/",
                        "props": {"message": "Hello"},
                        "version": "1",
                    }
                }
            )
        )
        assert 'id="my_app"' in rendered

    def test_data_page_valid_json(self):
        rendered = Template("{% load inertia_tags %} {% inertia %}").render(
            Context(
                {
                    "page__": {
                        "component": "Index",
                        "url": "/",
                        "props": {"message": "Hello"},
                        "version": "1",
                    }
                }
            )
        )

        # Assert that a JSON-parsable string is found in the data-page attr
        json_parsable_context_string = (
            '{"component": "Index", "url": "/", "props": {"message": "Hello"}, "version": "1"}'
        )
        assert f"data-page='{json_parsable_context_string}'" in rendered

        # Assert that the data-page attr value is a valid JSON string
        assert type(json.loads(json_parsable_context_string)) == dict



from django.http import HttpResponseRedirect

from .core import Inertia


class InertiaMiddleware:
    def __init__(self, get_response):
        self.get_response = get_response
        # One-time configuration and initialization.

    def __call__(self, request):
        # -> Code to be executed for each request before other middlewares and view are called.
        if not Inertia.get_version():
            Inertia.version(self.version(request))
        Inertia.share(self.share(request))

        response = self.get_response(request)

        # -> Code to be executed for each request/response after the view is called.
        response = self.check_version(request, response)
        # handle PUT, PATCH, DELETE redirections
        if (
            request.method in ["PUT", "PATCH", "DELETE"]
            and isinstance(response, HttpResponseRedirect)
            and response.status_code == 302
        ):
            response.status_code = 303
        return response

    def is_inertia(self, request):
        return request.headers.get("X-Inertia", False)

    def check_version(self, request, response):
        """In the event that the assets change, initiate a client-side location visit
        to force an update."""
        if (
            self.is_inertia(request)
            and request.method == "GET"
            and request.headers.get("X-Inertia-Version", "") != Inertia.get_version()
        ):
            # TODO: reflash session ?
            return Inertia.location(request.get_full_path_info())

        return response

    def version(self, request):
        """Determines the current asset version. Can be overriden."""
        # assets_url = config("inertia.public_path")
        # if assets_url:
        #     return hashlib.md5(assets_url.encode()).hexdigest()

        # manifest_file_path = public_path("mix-manifest.json")
        # if os.path.exists(manifest_file_path):
        #     hasher = hashlib.md5()
        #     with open(manifest_file_path, "rb") as manifest_file:
        #         buf = manifest_file.read()
        #         hasher.update(buf)
        #     return hasher.hexdigest()
        return 1

    def share(self, request):
        """Defines the props that are shared by default. Can be overriden."""
        errors = request.session.get("errors", False)
        success = request.session.get("success", False)
        # remove flash data for next request
        request.session["errors"] = False
        request.session["success"] = False
        return {"errors": Inertia.static(errors), "success": Inertia.static(success)}



import html
from inspect import signature
from typing import Any, Callable

from django.core.exceptions import ImproperlyConfigured
from django.http import HttpResponse, JsonResponse
from django.shortcuts import render
from dotty_dict import dotty

from .props import LazyProp, StaticProp
from .settings import settings


def load_callable_props(d, request):
    for k, v in d.items():
        if isinstance(v, dict):
            load_callable_props(v, request)
        elif callable(v):
            # evaluate prop and pass request if prop accept it
            if len(signature(v).parameters) > 0:
                d[k] = v(request)
            else:
                d[k] = v()
        elif isinstance(v, LazyProp):
            if len(signature(v.callable).parameters) > 0:
                d[k] = v(request)
            else:
                d[k] = v()
        elif isinstance(v, StaticProp):
            d[k] = v()


# Straightforward implementation of the Singleton Pattern
class Inertia(object):
    _instance = None
    shared_props = {}
    rendered_template = ""
    _version = ""

    def __new__(cls):
        if cls._instance is None:
            cls._instance = super(Inertia, cls).__new__(cls)
            # Put any initialization here.
            cls._instance.options = {
                "root_view": settings.INERTIA_ROOT_VIEW,
                "page_context": settings.INERTIA_PAGE_CONTEXT,
            }
            cls._instance.check_config()

        return cls._instance

    def check_config(self):
        if not self.options.get("root_view"):
            raise ImproperlyConfigured(
                "No Inertia template found. Either set INERTIA_ROOT_VIEW"
                "in settings.py or pass template parameter."
            )

    @classmethod
    def render(cls, request, component, props={}, view_data={}, custom_root_view=None):
        self = cls()
        page_data = self.get_page_data(request, component, props=props)

        if request.headers.get("X-Inertia", False):
            response = JsonResponse(page_data)
            response["X-Inertia"] = True
            response["Vary"] = "Accept"
            return response

        template = custom_root_view if custom_root_view else self.options.get("root_view")
        page_context = self.options.get('page_context')

        page_data = {page_context: page_data} if not view_data else \
            {**view_data, **{page_context: page_data}}

        return render(
            request,
            template,
            page_data,
        )

    @classmethod
    def location(cls, url):
        response = HttpResponse(status=409)
        response["X-Inertia-Location"] = url
        return response

    @staticmethod
    def lazy(callable: Callable):
        return LazyProp(callable)

    @staticmethod
    def static(value: Any):
        return StaticProp(value)

    def get_page_data(self, request, component, props):
        # merge shared props with page props, shared props keys takes precedence
        all_props = {**props, **self.get_shared_props()}
        # get props to use here if partial loading is requested
        props = self.get_props_to_use(request, all_props, component)
        # finally lazy load props and make request available to props being lazy loaded
        load_callable_props(props, request)

        page_data = {
            "component": self.get_component(component),
            "props": props,
            "url": request.get_full_path_info(),
            "version": self.get_version(),
        }

        return page_data

    def get_shared_props(self, key=None, default=None):
        """Get all Inertia shared props or the one with the given key."""
        if key:
            return dotty(self.shared_props, key, default)
        else:
            return self.shared_props

    @classmethod
    def version(cls, version):
        self = cls()
        self._version = version
        return self

    @classmethod
    def get_version(cls):
        self = cls()
        if callable(self._version):
            version = self._version()
        else:
            version = self._version
        return str(version)

    @classmethod
    def share(cls, key, value=None):
        self = cls()
        if isinstance(key, dict):
            self.shared_props = {**self.shared_props, **key}
        else:
            self.shared_props.update({key: value})
        return self

    def flush_shared(self):
        self.shared_props = {}

    def get_props_to_use(self, request, all_props, component):
        """Get props to return to the page:
        - when partial reload, required return 'only' props
        - add adapter props along view props (errors, message, auth ...)"""

        # partial reload feature
        only_props_header = request.headers.get("X-Inertia-Partial-Data")
        partial_component_header = request.headers.get("X-Inertia-Partial-Component") or {
            "name": ""
        }
        is_partial = only_props_header and partial_component_header == component
        props = {}

        if is_partial:
            only_props = only_props_header
            for key in all_props:
                # always load static props
                if key in only_props or isinstance(all_props[key], StaticProp):
                    props.update({key: all_props[key]})
        else:
            for prop_key, value in all_props.items():
                if not isinstance(value, LazyProp):
                    props.update({prop_key: value})

        return props

    def get_component(self, component):
        # TODO: check if escaping before here is needed
        return html.escape(component)



from django.core.handlers.wsgi import WSGIRequest
from django.test import RequestFactory
from django.test.testcases import SimpleTestCase

from django_inertia import Inertia


class TestLazyAndPartialReloads(SimpleTestCase):
    def setUp(self):
        factory = RequestFactory()
        self.request = factory.get("/test-inertia")
        # for idempotent tests
        Inertia().flush_shared()

    def test_callable_prop_is_included_initially(self):
        props = {"normal": "value", "lazy": lambda r: "lazy_value"}
        response = Inertia.render(self.request, "Index", props)
        self.assertContains(response, "normal")
        self.assertContains(response, "lazy")

        self.request.headers = {
            "X-Inertia-Partial-Data": ["normal"],
            "X-Inertia-Partial-Component": "Index",
        }
        response = Inertia.render(self.request, "Index", props)
        self.assertContains(response, "normal")
        self.assertNotContains(response, "lazy")

    def test_lazy_prop_is_not_include_initially(self):
        props = {"normal": "value", "lazy": Inertia.lazy(lambda r: "lazy_value")}
        response = Inertia.render(self.request, "Index", props)
        self.assertContains(response, "normal")
        self.assertNotContains(response, "lazy")

        self.request.headers = {
            "X-Inertia-Partial-Data": ["lazy"],
            "X-Inertia-Partial-Component": "Index",
        }
        response = Inertia.render(self.request, "Index", props)
        self.assertContains(response, "lazy")
        self.assertNotContains(response, "normal")

    def test_lazy_or_callable_prop_receives_request(self):
        def my_callable(req):
            assert isinstance(req, WSGIRequest)
            return req.get_full_path()

        props = {"callable_prop": my_callable, "lazy_prop": Inertia.lazy(my_callable)}
        response = Inertia.render(self.request, "Index", props)
        self.assertContains(response, "/test-inertia")

    def test_can_share_globally_lazy_or_callable_props(self):
        Inertia.share(
            {"user": Inertia.lazy(lambda r: {"first_name": "John"}), "notifs": lambda r: []}
        )
        response = Inertia.render(self.request, "Index")
        self.assertNotContains(response, "user")
        self.assertContains(response, "notifs")

        self.request.headers = {
            "X-Inertia-Partial-Data": ["user", "notifs"],
            "X-Inertia-Partial-Component": "Index",
        }
        response = Inertia.render(self.request, "Index")
        self.assertContains(response, "user")
        self.assertContains(response, "notifs")

    def test_that_static_props_are_always_loaded(self):
        Inertia.share("errors", Inertia.static("Unknown error"))
        response = Inertia.render(self.request, "Index")
        self.assertContains(response, "errors")

        self.request.headers = {
            "X-Inertia-Partial-Data": ["message"],
            "X-Inertia-Partial-Component": "Index",
        }
        response = Inertia.render(self.request, "Index", {"message": "Hello"})
        self.assertContains(response, "message")
        self.assertContains(response, "errors")



from django.conf import settings as django_settings

__all__ = ["settings"]


class LazySettings:
    INERTIA_ROOT_VIEW = "app.html"
    INERTIA_PAGE_CONTEXT = "page__"

    def __getattribute__(self, name):
        try:
            return getattr(django_settings, name)
        except AttributeError:
            return super().__getattribute__(name)


settings = LazySettings()



from setuptools import find_packages, setup

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name="django-inertia",
    version="1.3.0",
    packages=find_packages(),
    description="Server-side Django adapter for Inertia.js",
    long_description=long_description,
    long_description_content_type="text/markdown",
    # The project's main homepage.
    url="https://github.com/girardinsamuel/django-inertia",
    # Author details
    author="Samuel Girardin",
    author_email="samuelgirardin@pm.me",
    # Choose your license
    license="MIT license",
    # If your package should include things you specify in your MANIFEST.in file
    # Use this option if your package needs to include files that are not python files
    # like html templates or css files
    include_package_data=True,
    # See https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Topic :: Software Development :: Libraries :: Application Frameworks",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Framework :: Django",
    ],
    # What does your project relate to?
    keywords="Django, Python, Inertia.js",
    # List run-time dependencies here.  These will be installed by pip when
    # your project is installed. For an analysis of "install_requires" vs pip's
    # requirements files see:
    # https://packaging.python.org/en/latest/requirements.html
    install_requires=[
        "django>=3,<5",
        "dotty-dict==1.3.0",
    ],
    python_requires=">=3.7.0",
    # List additional groups of dependencies here (e.g. development
    # dependencies). You can install these using the following syntax,
    # for example:
    # $ pip install -e .[dev,test]
    # $ pip install your-package[dev,test]
    extras_require={
        "dev": [
            "black",
            "flake8",
            "flake8-isort",
            "twine",
            "wheel",
            "pytest",
            "pytest-django",
            "pytest-cov",
            "coverage",
            "bump2version",
        ],
    },
)



from .core import Inertia


def assert_component(response, component):
    assert (
        response.context[Inertia.options.get('page_context')]["component"] == component
    ), f"Asserted {component}, " \
       f"got {response.context[Inertia.options.get('page_context')]['component']}"


def assert_props(response, key, value=None):
    props = response.context[Inertia.options.get("page_context")]["props"]
    assert key in props
    if value:
        assert props.get(key, None) == value
